use std::time::Duration;

use once_cell::sync::Lazy;
use rustracing_jaeger::Span;
use tiny_http::{Request, Response, ResponseBox};

use crate::{
    config::CONFIG,
    db::DBConnection,
    metrics,
    routes::{header, AppError}
};

#[derive(serde::Deserialize, Clone, Debug)]
pub struct GitlabTokens {
    pub access_token: String,
    pub refresh_token: String
}

#[derive(serde::Deserialize, Clone, Debug)]
pub struct GitlabUser {
    pub username: String,
    pub is_admin: bool
}

pub static REDIRECT_URL: Lazy<String> = Lazy::new(|| CONFIG.gitlab_redirect_url.clone() + "/api/auth/gitlab_callback");
pub static TOKEN_URL: Lazy<String> = Lazy::new(|| format!("{}/login/oauth/access_token", CONFIG.gitlab_api_url.clone()));
pub static USER_URL: Lazy<String> = Lazy::new(|| format!("{}/api/v1/user", CONFIG.gitlab_api_url.clone()));
pub static AUTHORIZE_URL: Lazy<String> = Lazy::new(|| format!("{}/login/oauth/authorize", CONFIG.gitlab_url.clone()));

pub fn get_gitlab_token(span: &Span, code_or_token: String, token: bool) -> Option<GitlabTokens> {
    let _span = metrics::span("get_gitlab_token", span);
    let grant_type = match token {
        true => "refresh_token",
        false => "authorization_code"
    };
    let code_name = match token {
        true => "refresh_token",
        false => "code"
    };
    ureq::post(&TOKEN_URL)
        .send_json(ureq::json! ({
            "grant_type": grant_type,
            "client_id": &CONFIG.gitlab_id,
            "client_secret": &CONFIG.gitlab_secret,
            "redirect_uri": *REDIRECT_URL,
            code_name: code_or_token
        })).ok()?.into_json().ok()
}

pub fn get_gitlab_user(span: &Span, token: String) -> Option<GitlabUser> {
    static CACHE: Lazy<stretto::Cache<String, GitlabUser>> =
        Lazy::new(|| stretto::CacheBuilder::new(1000, 100).finalize().expect("Failed to create gitlab cache"));
    match CACHE.get(&token) {
        None => {
            let _span = metrics::span("get_gitlab_user", span);
            ureq::get(&USER_URL)
                .set("Authorization", &format!("Bearer {}", token))
                .call()
                .ok()?
                .into_json::<GitlabUser>()
                .ok()
                .map(|v| {
                    CACHE.insert_with_ttl(token, v.clone(), 1, Duration::from_secs(500));
                    v
                })
        }
        Some(v) => Some(v.value().clone())
    }
}

pub fn gitlab(_: &Span, _: &mut Request, _: &mut DBConnection) -> Result<ResponseBox, AppError> {
    let uri = format!(
        "{}?redirect_uri={}&client_id={}&scope=read_user&response_type=code",
        AUTHORIZE_URL.as_str(),
        REDIRECT_URL.as_str(),
        CONFIG.gitlab_id
    );
    Ok(Response::empty(302).with_header(header("location", &uri)).boxed())
}

pub fn gitlab_callback(
    span: &Span,
    _: &mut Request,
    db: &mut DBConnection,
    query_string: &str
) -> Result<ResponseBox, AppError> {
    use crate::db::UserRole;

    let code: &str = query_string
        // a=b&code=c&d
        .split('&')
        // ['a=b', 'code=c', 'd']
        .find_map(|prop|
            prop.split_once('=')
                // [Some(('a', 'b')), Some(('code', 'c')), None]
                .and_then(|v|
                    v.0.eq("code")
                        .then_some(v.1)
                )
        )
        .ok_or(AppError::BadRequest("Query code missing"))?;

    let tokens = get_gitlab_token(span, code.to_string(), false).ok_or(AppError::Unauthorized("Invalid code"))?;
    let gitlab_user =
        get_gitlab_user(span, tokens.access_token.clone()).ok_or(AppError::Unauthorized("Invalid code"))?;

    let user = db.find_user(span, &gitlab_user.username, true);

    let user = match user {
        Some(mut v) => {
            v.gitlab_at = Some(tokens.access_token);
            v.gitlab_rt = Some(tokens.refresh_token);
            db.save_user(span, &v);
            v
        }
        None => db.create_user_gitlab(
            span,
            gitlab_user.username,
            if gitlab_user.is_admin {
                UserRole::Admin
            } else {
                UserRole::Disabled
            },
            tokens.access_token,
            tokens.refresh_token
        )
    };

    if user.role == UserRole::Disabled {
        Ok(
            Response::from_data("<!DOCTYPE html><html><h2>Your account is disabled, please contact an admin.<br/><a href=\"/login\">Go to login page</a></h2></html>")
                .with_header(header("content-type", "text/html; charset=utf-8"))
                .boxed()
        )
    } else {
        let uri = format!("/set_token?token={}", super::get_token(span, &user, db));
        Ok(Response::empty(302).with_header(header("location", &uri)).boxed())
    }
}
